%% MAIN
close all
clear all
clc

addpath('functions','-end');
addpath('system_models','-end');


x_init = [1; 1; pi/2];


%% ==================================================================
%  ====================== Trajectory generator ======================
%  ==================================================================

% System model: point mass kinematic
T_sys = 0.01;
[pred_model_traj] = point_kinematic_model(T_sys);
pred_model_traj.x0 = [x_init(1); x_init(2); 0; 0];
pred_model_traj.v_des = [0; 0];
pred_model_traj.ref = [18; 9; 0; 0];




%% ==================================================================
%  ============================ Planning ============================
%  ==================================================================

% System model: point mass kinematic
T_plan = 0.1;
[pred_model_plan] = point_kinematic_model(T_plan);

% Constraints
x_min = [-inf; -inf; -0.5; -0.5];
x_max = [inf; inf; 0.5; 0.5];
u_min = [-0.1; -0.1];
u_max = [0.1; 0.1];
du_min = [-0.1; -0.1];
du_max = [0.1; 0.1];

% Obstacle avoidance
r_act = 2;
r_safe = 0.25;
[map] = obs_design(r_act, r_safe);

% Waypoints definition
r_tol = 0.2;
[wp] = waypoint_design(r_tol);

% MPC path planning
mpc_plan_param.Ts = T_plan;
mpc_plan_param.Np = 30;
mpc_plan_param.Nc = 30;
mpc_plan_param.Q = diag([1 1 1 1]);
mpc_plan_param.R = diag([10 10]);
mpc_plan_param.A = pred_model_plan.A;
mpc_plan_param.B = pred_model_plan.B;
mpc_plan_param.C = pred_model_plan.C;
mpc_plan_param.nx = size(pred_model_plan.A,1);
mpc_plan_param.nu = size(pred_model_plan.B,2);
mpc_plan_param.ny = size(pred_model_plan.C,1);
mpc_plan_param.x_c = [x_min, x_max];
mpc_plan_param.u_c = [u_min, u_max];
mpc_plan_param.du_c = [du_min, du_max];

mpc_plan = mpc_design(mpc_plan_param);




%% =================================================================
%  ============================ Control ============================
%  =================================================================

T_ctrl = 0.2;
x0 = [x_init(1); x_init(2);x_init(3)];

% Constraints
x_min = [-inf; -inf; -inf];
x_max = [inf; inf; inf];
u_min = [-0.5];
u_max = [0.5];
du_min = [-0.5];
du_max = [0.5];


% MPC control
mpc_ctrl_param.Ts = T_ctrl;
mpc_ctrl_param.nx = 3;
mpc_ctrl_param.nu = 1;
mpc_ctrl_param.ny = 2;
mpc_ctrl_param.Np = 25;
mpc_ctrl_param.Nc = 10;
mpc_ctrl_param.Q = diag([1 1]);
mpc_ctrl_param.R = diag([1]);
mpc_ctrl_param.x_c = [x_min, x_max];
mpc_ctrl_param.u_c = [u_min, u_max];
mpc_ctrl_param.du_c = [du_min, du_max];





%% ==================================================================
%  =========================== Simulation ===========================
%  ==================================================================
t_sim = inf;

% open('two_layer_mpc_sim');

tic;
sim('two_layer_mpc_sim');
toc;




%% ==================================================================
%  =========================== Plot ===========================
%  ==================================================================


%% Plot map and waypoints

% map
figure(1); hold on; axis equal;
for i=1:length(map.Sx)
    p1 = plot([map.Sx(i,1);map.Sx(i,1)], [map.Sx(i,2);map.Sx(i,3)], 'k');
    set(get(get(p1, 'Annotation'), 'LegendInformation'), 'IconDisplayStyle', 'off');
end
for i=1:length(map.Sy)
    p2 = plot([map.Sy(i,2);map.Sy(i,3)], [map.Sy(i,1);map.Sy(i,1)], 'k');
    set(get(get(p2, 'Annotation'), 'LegendInformation'), 'IconDisplayStyle', 'off');
end 
xlabel('x [m]');
ylabel('y [m]');


% waypoints
for i=1:length(wp.points)
    p3 = plot(wp.points(1,i), wp.points(2,i), 'kx');
    set(get(get(p3, 'Annotation'), 'LegendInformation'), 'IconDisplayStyle', 'off');
end




%% Planning trajectory 

figure(1); hold on;
title('Trajectory');

% initial pose
p4 = plot(pred_model_traj.x0(1), pred_model_traj.x0(2), '^r');
set(get(get(p4, 'Annotation'), 'LegendInformation'), 'IconDisplayStyle', 'off');

% target
p5 = plot(pred_model_traj.ref(1), pred_model_traj.ref(2), 'or');
set(get(get(p5, 'Annotation'), 'LegendInformation'), 'IconDisplayStyle', 'off');

% trajectories
plot(x_gen.data(:,1), x_gen.data(:,2), 'b', 'LineStyle', '-', 'LineWidth', 1, 'DisplayName', 'reference');
plot(x.data(:,1), x.data(:,2), 'r', 'LineStyle', '-', 'LineWidth', 1, 'DisplayName', 'trajectory');



%% Planning 

% Position
figure(2); hold on; grid on;
title('Position');

subplot(3,1,1); hold on; grid on;
title('X');
plot(trajectory.time, trajectory.data(:,1), 'b', 'LineWidth', 1, 'DisplayName', 'ref');

subplot(3,1,2); hold on; grid on;
title('Y');
plot(trajectory.time, trajectory.data(:,2), 'b', 'LineWidth', 1, 'DisplayName', 'ref');

subplot(3,1,3); hold on; grid on;
title('\psi');
plot(trajectory.time, trajectory.data(:,2)*180/pi, 'b', 'LineWidth', 1, 'DisplayName', 'ref');


% input
figure(3); hold on; grid on;
title('Input');

subplot(3,1,1); hold on; grid on;
title('a_x');
plot(u_plan.time, u_plan.data(:,1), 'b', 'LineWidth', 1, 'DisplayName', 'ref');

subplot(3,1,2); hold on; grid on;
title('a_y');
plot(u_plan.time, u_plan.data(:,2), 'b', 'LineWidth', 1, 'DisplayName', 'ref');

subplot(3,1,3); hold on; grid on;
title('a');
plot(u_plan_body.time, u_plan_body.data(:,1), 'b', 'LineWidth', 1, 'DisplayName', 'ref');



%% Control 

% State
figure(4); hold on; grid on;
title('Position');

subplot(3,1,1); hold on; grid on;
title('X');
plot(x.time, x.data(:,1), 'r', 'LineWidth', 1, 'DisplayName', 'ref');

subplot(3,1,2); hold on; grid on;
title('Y');
plot(x.time, x.data(:,2), 'r', 'LineWidth', 1, 'DisplayName', 'ref');

subplot(3,1,3); hold on; grid on;
title('\psi');
plot(x.time, x.data(:,3)*180/pi, 'r', 'LineWidth', 1, 'DisplayName', 'ref');


%% Input
figure(5); hold on; grid on;
title('\beta');
plot(u_ctrl.time, u_ctrl.data(:,1)*180/pi, 'r', 'LineWidth', 1, 'DisplayName', '\beta');
xlabel('x [t]');
ylabel('y [°/s]');







return


%% Save data
% save('reference.mat', 'trajectory')
