%% MAIN
close all
clear all
clc

addpath('../00 - Functions/mpc','-end')
addpath('../00 - Functions/plant_modelling','-end')
addpath('../00 - Functions/plot','-end')
addpath('../00 - Functions/utils','-end')


load('trajectory.mat')
trajectory_ref = X_plant(4:5,:);
x0 = X_plant(:,1);
REF = X_MPC;
t_ref = t_fin;
clear X_plant X_MPC U_MPC t_fin



%% =================================================================
%  ============================ Plant ============================
%  =================================================================

T = 0.01;
X_plant = x0;

% vehicle data
sys_model.lf = 0.75;
sys_model.lr = 0.75;
sys_model.L = sys_model.lf+sys_model.lr;
sys_model.W = 0.8;
sys_model.m = 800;
sys_model.Iz = sys_model.m*(sys_model.L^2 + sys_model.W^2)/12;
sys_model.Cf = 10000;
sys_model.Cr = 10000;



%% =================================================================
%  ============================ Track ==============================
%  =================================================================

track_n = 6;
W_track = 1;
[PointAndTangent, x0_track] = track_fun(track_n, W_track);




%% ==================================================================
%  ======================= MPC initialization =======================
%  ==================================================================

X_MPC = REF(:,1);

% state constraints [vx, vy, wz, s, ey, epsi]
x_c = [0,           +3
       -0.5,        +0.5
       -40*pi/180,  +40*pi/180
       0,           +100
       -W_track,    +W_track
       -50*pi/180,  +50*pi/180];

% input constraints [delta_f, a]
u_c = [-30*pi/180,  +30*pi/180
       -0.5,          +0.1];

% input rate constraints [delta_f, a]
du_c = [-5*pi/180,     +5*pi/180
        -0.005,           +0.005];
    
    
    

%% ==================================================================
%  ============================== MPC ===============================
%  ==================================================================

T_ctrl = 0.05;
method = 0;
lpv = false;

% MPC tuning
Np = 10;
Nc = Np;
Q = diag([10 1 1 1 100 1]);
R = diag([1 1]);
S = diag([0 0]);

U_old = zeros(Nc*2,1);
t_sim = 50;

k = 1;
flag = 0;

tic;
while (flag == 0)
    
    clc
    t = T_ctrl*k;
    
    %% Print info
    fprintf('Sample: k = %d, time: t = %.2f\n', [k, t]);
    fprintf('Np = %d, Nc = %d \n', [Np, Nc]);
    fprintf('Velocity =         [vx = %.2fm/s,  vy = %.2fm/s,   wz =   %.1f°/s] \n', [X_MPC(1,k), X_MPC(2,k), X_MPC(3,k)*180/pi]);
    fprintf('Curvilinear pose = [s =  %.2fm,    ey = %.2fm,     epsi = %.1f°] \n',   [X_MPC(4,k), X_MPC(5,k), X_MPC(6,k)*180/pi]);

    %% MPC
    U_star = adaptive_mpc(X_MPC(:,k), REF(:,k), U_old, T_ctrl, Np, Nc, Q, R, S, x_c, u_c, du_c, PointAndTangent, method, lpv);
    if (isempty(U_star) | isnan(U_star))
        u_star = [];
        fprintf('Problem infeasible at k = %d \n', k);
        t_fin = T_ctrl*(k-1);
        break;
    end
    U_old = U_star;
    U_MPC(:,k) = U_star(1:2);
    fprintf('Input =            [delta_f = %.1f°,  a = %.2fm/s^2]\n', [U_MPC(1,k)*180/pi, U_MPC(2,k)]);

    %% Plant
    x = X_plant(:,k);
    u = U_MPC(:,k);
    for i=1:T_ctrl/T
        [x_plus, alpha] = bicycle_dynamics_discrete(x, u, sys_model, T);
        x = x_plus;
    end    
    X_plant(:,k+1) = x;
    
    
    %% Update sample
    k = k+1;
    u_old = U_MPC(:,k-1);
    
    
    %% Conversion from global to curvilinear coordinates
    [s, ey, epsi, CompletedFlag] = global2local_fun(X_plant(4:6,k), x0_track, PointAndTangent, W_track);
    X_MPC(:,k) = [X_plant(1,k); X_plant(2,k); X_plant(3,k); s; ey; epsi];
    
    
    if CompletedFlag == 0
        flag = 1;
        t_fin = T_nav*(k);
        break;
    end
    
    %% Exit condition
    if (abs(X_MPC(4,k) - REF(4,end)) <= 10e-2)
        fprintf('Terget reached \n');
        flag = 1;
        t_fin = T_ctrl*(k);
    elseif REF(:,k) == REF(:,end)
        REF(:,k+1) = REF(:,end);
    elseif (t == t_sim)
        fprintf('Simulation time reached \n');
        flag = 1;
        t_fin = T_ctrl*(k);
    end
    
    
    
end
toc;


%% ==================================================================
%  =========================== Plot =================================
%  ==================================================================

t_vec = linspace(0, t_fin, (t_fin)/T_ctrl);

figure(1); hold on; grid on;
title('Trajectory');
plot(trajectory_ref(1,:), trajectory_ref(2,:), 'r');
plot(X_plant(4,1:10:end), X_plant(5,1:10:end), 'xb', 'LineWidth', 0.5);
arrow(X_plant(4,1:10:end), X_plant(5,1:10:end), X_plant(6,1:10:end), 'b');



% UGV curvilinear
figure(3); 
title('Curvilinear states');
m = 3; n = 1;

subplot(m, n, 1); hold on; grid on;
title('s');
ylim([x_c(4,1)-0.5, x_c(4,2)+0.5]);
plot(t_vec, X_MPC(4,1:end), 'k', 'DisplayName', 's');
plot(t_vec, REF(4,1:end), 'r', 'DisplayName', 's_REF');

subplot(m, n, 2); hold on; grid on;
title('e_y');
ylim([x_c(5,1)-x_c(5,1)*10/100, x_c(5,2)+x_c(5,2)*10/100]);
plot(t_vec, X_MPC(5,1:end), 'k', 'DisplayName', 'e_y');
plot(t_vec, REF(5,1:end), 'r', 'DisplayName', 'ey_REF');

subplot(m, n, 3); hold on; grid on;
title('e_{psi}');
ylim([x_c(6,1)*180/pi+x_c(6,1)*180/pi*10/100, x_c(6,2)*180/pi+x_c(6,2)*180/pi*10/100]);
plot(t_vec, X_MPC(6,1:end)*180/pi, 'k', 'DisplayName', 'e_{psi}');
plot(t_vec, REF(6,1:end)*180/pi, 'r', 'DisplayName', 'epsi_REF');




% UGV velocity
figure(4); hold on; grid on;
title('Velocity');
m = 3; n = 1;

subplot(m, n, 1); hold on; grid on;
title('v_x');
ylim([x_c(1,1)+x_c(1,1)*10/100, x_c(1,2)+x_c(1,2)*10/100]);
plot(t_vec, X_MPC(1,1:end), 'k', 'DisplayName', 'vx');
plot(t_vec, REF(1,1:end), 'r', 'DisplayName', 'vx_REF');

subplot(m, n, 2); hold on; grid on;
title('v_y');
ylim([x_c(2,1)+x_c(2,1)*10/100, x_c(2,2)+x_c(2,2)*10/100]);
plot(t_vec, X_MPC(2,1:end), 'k', 'DisplayName', 'vy');
plot(t_vec, REF(2,1:end), 'r', 'DisplayName', 'vy_REF');

subplot(m, n, 3); hold on; grid on;
title('w_z');
ylim([x_c(3,1)*180/pi+x_c(3,1)*180/pi*10/100, x_c(3,2)*180/pi+x_c(3,2)*180/pi*10/100]);
plot(t_vec, X_MPC(3,1:end)*180/pi, 'k', 'DisplayName', 'wz');
plot(t_vec, REF(3,1:end)*180/pi, 'r', 'DisplayName', 'wz_REF');



% UGV input
figure(5); hold on; grid on;
title('Input');
m = 2; n = 1;
subplot(m, n, 1); hold on; grid on;
title('delta_f');
ylim([u_c(1,1)*180/pi+u_c(1,1)*180/pi*10/100, u_c(1,2)*180/pi+u_c(1,2)*180/pi*10/100]);
plot(t_vec(1:end-1), U_MPC(1,1:end)*180/pi, 'k', 'DisplayName', 'delta_f');

subplot(m, n, 2); hold on; grid on;
title('a');
ylim([u_c(2,1)+u_c(2,1)*10/100 u_c(2,2)+u_c(2,2)*10/100]);
plot(t_vec(1:end-1), U_MPC(2,1:end), 'k', 'DisplayName', 'a');

