%% MAIN
close all
clear all
clc

addpath('../00 - Functions/mpc','-end')
addpath('../00 - Functions/plant_modelling','-end')
addpath('../00 - Functions/plot','-end')
addpath('../00 - Functions/utils','-end')



%% =================================================================
%  ============================ Plant ============================
%  =================================================================

% UGV = [vx, vy, wz, X, Y, psi];
x0 = [1; 0; 0; 0; 0; 90*pi/180];
T = 0.001;

figure(1); hold on; grid on;
title('Trajectory');
plot(x0(4), x0(5), 'or');
arrow(x0(4), x0(5), x0(6), 'r');

% vehicle data
sys_model.lf = 0.75;
sys_model.lr = 0.75;
sys_model.L = sys_model.lf+sys_model.lr;
sys_model.W = 0.8;
sys_model.m = 800;
sys_model.Iz = sys_model.m*(sys_model.L^2 + sys_model.W^2)/12;
sys_model.Cf = 10000;
sys_model.Cr = 10000;



%% =================================================================
%  ============================ Track ==============================
%  =================================================================

track_n = 1;
W_track = 5;
[PointAndTangent, x0_track] = track_fun(track_n, W_track);



%% ==================================================================
%  ======================= MPC initialization =======================
%  ==================================================================


% initial condition
X_plant = x0;
[s0, ey0, epsi0] = global2local_fun(x0(4:6), x0_track, PointAndTangent, W_track);
X_MPC = [x0(1); x0(2); x0(3); s0; ey0; epsi0];
% X_MPC = [0.1; 0; 0; 0.1; 0; 0];


% target
target = [PointAndTangent(end,1); PointAndTangent(end,2); PointAndTangent(end,3)];
[s_target, ey_target, epsi_target] = global2local_fun(target, x0(4:6), PointAndTangent, W_track);
% ref = [1.5; 0; 0; s_target-s0; ey_target; epsi_target];
ref = [2; 0; 0; 0; 0; 0];


% state constraints [vx, vy, wz, s, ey, epsi]
x_c = [-0.2,           +0.2
       -0.5,           +0.5
       -50*pi/180,  +50*pi/180
       0,           +1
       -W_track,          +W_track
       -50*pi/180,  +50*pi/180];

% input constraints [delta_f, a]
u_c = [-30*pi/180,  +30*pi/180
       -0.1,          +0.1];

% input rate constraints [delta_f, a]
du_c = [-1*pi/180,     +1*pi/180
        -0.05,           +0.05];



%% ==================================================================
%  ============================== MPC ===============================
%  ==================================================================

T_nav = 0.02;
method = 0;
lpv = false;

% MPC tuning
Np = 20;
Nc = Np;
Q = diag([1 1 1 1 10 1]);
R = diag([1 1]);
S = diag([0 0]);

k = 1;
flag = 0;
t_sim = 20;

U_old = zeros(Nc*2,1);
U_MPC = [0; 0];
J = 0;

while (flag == 0)
    
    clc
    t = T_nav*k;
    
    %% Print info
    fprintf('Sample: k = %d, time: t = %.2f\n', [k, t]);
    fprintf('Np = %d, Nc = %d \n', [Np, Nc]);
    fprintf('Velocity =         [vx = %.2fm/s,  vy = %.2fm/s,   wz =   %.1f°/s] \n', [X_MPC(1,k), X_MPC(2,k), X_MPC(3,k)*180/pi]);
    fprintf('Curvilinear pose = [s =  %.2fm,    ey = %.2fm,     epsi = %.1f°] \n',   [X_MPC(4,k), X_MPC(5,k), X_MPC(6,k)*180/pi]);
    fprintf('Cost = [J =  %.2f] \n',   [J]);


    %% MPC
    U_star = adaptive_mpc(X_MPC(:,k), X_MPC(:,1), ref, U_old, U_MPC, T_nav, Np, Nc, Q, R, S, x_c, u_c, du_c, PointAndTangent, method, lpv);
%     U_star = [20*cos(t)*pi/180; 0];
    if (isempty(U_star) | isnan(U_star))
        u_star = [];
        fprintf('Problem infeasible at k = %d \n', k);
        t_fin = T_nav*(k-1);
        break;
    end
    U_old = U_star;
    U_MPC(:,k) = U_star(1:2);

    fprintf('Input =            [delta_f = %.1f°,  a = %.2fm/s^2]\n', [U_MPC(1,k)*180/pi, U_MPC(2,k)]);
    if not(U_MPC(1,k)<=0.1*pi/180)
        U_MPC(1,k)
    end

    %% Plant
    x = X_plant(:,k);
    u = U_MPC(:,k);
%     for i=1:T_nav/T
%         [x_plus, alpha] = bicycle_dynamics_discrete(x, u, sys_model, T);
%         x = x_plus;
%     end
    
	[t_x, x_plus] = ode23(@(t,x)bicycle_dynamics_continuous(x, u, sys_model), linspace(0, T_nav, T_nav/T), x);
    x = x_plus(end,:);
    
    X_plant(:,k+1) = x;
%     alpha_f(k) = alpha;
    

    %% Cost
    J = J + X_MPC(:,k)'*Q*X_MPC(:,k) + U_MPC(:,k)'*R*U_MPC(:,k);
    
    
    %% Update sample
    k = k+1;
    u_old = U_MPC(:,k-1);
    
%     figure(1); hold on; grid on; axis equal;
%     p2 = plot(X_plant(4,k), X_plant(5,k), 'or', 'LineWidth', 0.5);
%     arrow(X_plant(4,k), X_plant(5,k), X_plant(6,k), 'b');
    
%     delete(p1)
    
    
    %% Conversion from global to curvilinear coordinates
    [s, ey, epsi, CompletedFlag] = global2local_fun(X_plant(4:6,k), x0_track, PointAndTangent, W_track);
    X_MPC(:,k) = [X_plant(1,k); X_plant(2,k); X_plant(3,k); s; ey; epsi];
    
    if CompletedFlag == 0
        flag = 1;
        t_fin = T_nav*(k);
        break;
    end
    

    
    %% Exit condition
    if (abs(X_MPC(4,k) - s_target) <= 10e-2)
        fprintf('Terget reached \n');
        flag = 1;
        t_fin = T_nav*(k-1);
    elseif (t == t_sim)
        fprintf('Simulation time reached \n');
        flag = 1;
        t_fin = T_nav*(k-1);
    elseif k>147
        k;
    end
        
end



%% ==================================================================
%  =========================== Plot =================================
%  ==================================================================

t_vec = linspace(0, t_fin, (t_fin+T_nav)/T_nav);

figure(1); hold on; grid on; axis equal;
plot(X_plant(4,:), X_plant(5,:), '-r', 'LineWidth', 0.5);
arrow(X_plant(4,end), X_plant(5,end), X_plant(6,end), 'b');

% UGV position
figure(2); hold on; grid on;
title('Position');
m = 3; n = 1;

subplot(m, n, 1); hold on; grid on; 
title('x');
plot(t_vec, X_plant(4,1:end), 'k', 'DisplayName', 'x');

subplot(m, n, 2); hold on; grid on; 
title('y');
plot(t_vec, X_plant(5,1:end), 'k', 'DisplayName', 'y');

subplot(m, n, 3); hold on; grid on; 
title('psi');
plot(t_vec, X_plant(6,1:end), 'k', 'DisplayName', 'psi');


% UGV curvilinear
figure(3); 
title('Curvilinear states');
m = 3; n = 1;

subplot(m, n, 1); hold on; grid on;
title('s');
ylim([x_c(4,1)-0.5, x_c(4,2)+0.5]);
plot(t_vec, X_MPC(4,1:end), 'k', 'DisplayName', 's');
yline(x_c(4,1), 'r--'); yline(x_c(4,2), 'r--');

subplot(m, n, 2); hold on; grid on;
title('e_y');
ylim([x_c(5,1)-x_c(5,1)*10/100, x_c(5,2)+x_c(5,2)*10/100]);
plot(t_vec, X_MPC(5,1:end), 'k', 'DisplayName', 'e_y');
yline(x_c(5,1), 'r--'); yline(x_c(5,2), 'r--');

subplot(m, n, 3); hold on; grid on;
title('e_{psi}');
ylim([x_c(6,1)*180/pi+x_c(6,1)*180/pi*10/100, x_c(6,2)*180/pi+x_c(6,2)*180/pi*10/100]);
plot(t_vec, X_MPC(6,1:end)*180/pi, 'k', 'DisplayName', 'e_{psi}');
yline(x_c(6,1)*180/pi, 'r--'); yline(x_c(6,2)*180/pi, 'r--');


% UGV velocity
figure(4); hold on; grid on;
title('Velocity');
m = 3; n = 1;

subplot(m, n, 1); hold on; grid on;
title('v_x');
plot(t_vec, X_MPC(1,1:end), 'k', 'DisplayName', 'v_x');

subplot(m, n, 2); hold on; grid on;
title('v_y');
plot(t_vec, X_MPC(2,1:end), 'k', 'DisplayName', 'v_y');

subplot(m, n, 3); hold on; grid on;
title('w_z');
plot(t_vec, X_MPC(3,1:end)*180/pi, 'k', 'DisplayName', 'w_z');


% UGV input
figure(5); hold on; grid on;
title('Input');
m = 2; n = 1;
subplot(m, n, 1); hold on; grid on;
title('delta_f');
ylim([u_c(1,1)*180/pi+u_c(1,1)*180/pi*10/100, u_c(1,2)*180/pi+u_c(1,2)*180/pi*10/100]);
plot(t_vec(1:end-1), U_MPC(1,1:end)*180/pi, 'k', 'DisplayName', 'delta_f');
% plot(t_vec(1:end-1), alpha_f(1:end)*180/pi, 'b', 'DisplayName', 'alpha_f');
yline(u_c(1,1)*180/pi, 'r--'); yline(u_c(1,2)*180/pi, 'r--');

subplot(m, n, 2); hold on; grid on;
title('a');
ylim([u_c(2,1)+u_c(2,1)*10/100 u_c(2,2)+u_c(2,2)*10/100]);
plot(t_vec(1:end-1), U_MPC(2,1:end), 'k', 'DisplayName', 'a');
yline(u_c(2,1), 'r--'); yline(u_c(2,2), 'r--');


return

%% Save plots
prompt = 'Do you want to save the plots? y/n [n]: ';
str = input(prompt,'s');
if str == 'y'
    
    save('data/trajectory.mat', 'X_MPC', 'X_plant', 'U_MPC', 't_fin', 'T_nav')
    
    % mpc params
    mpc.Np = Np;
    mpc.Nc = Nc;
    mpc.Q = Q;
    mpc.R = R;
    save data/mpc_data.mat mpc sys_model
    
    % trajectory plot
    filename = sprintf('data/trajectory.png');
    exportgraphics(figure(1), filename);
    
    % trajectory plot
    filename = sprintf('data/velocities.png');
    exportgraphics(figure(4), filename);
    
    % trajectory plot
    filename = sprintf('data/curvliniear_coordinates.png');
    exportgraphics(figure(3), filename);
    
    % trajectory plot
    filename = sprintf('data/input.png');
    exportgraphics(figure(5), filename);
    
end

close all


